/*	$Id: pomEuler.c 2010/06/02 18:30:17 mallet $ */

/* 
 * Copyright (c) 2007,2009-2010 LAAS/CNRS                  --  Tue Aug 28 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include "pomEuler.h"


const POM_EULER_V pomEulerVIdentity = {
  {
    /*          O    P    R     X    Y    Z   */
    /*      */ 0.0, 0.0, 0.0,  0.0, 0.0, 0.0 },
  {
    /*          0    */
    /*  O  */  0.0,

    /*          1    2   */
    /*  P  */  0.0, 0.0,

    /*          3    4   5  */
    /*  R  */  0.0, 0.0, 0.0,

    /*          6    7   8     9  */
    /*  X  */  0.0, 0.0, 0.0,  0.0,

    /*         10   11   12    13   14  */
    /*  Y  */  0.0, 0.0, 0.0,  0.0, 0.0,

    /*         15   16   17    18   19   20   */
    /*  Z  */  0.0, 0.0, 0.0,  0.0, 0.0, 0.0
}};

/* ----------------------------------------------------------------------
 *
 * pomEulerPrintMat - print matrix (mainly debug purpose)
 *
 */
static void pomEulerPrintMat(char *msg, double mat[16]) 
{
  int i, j;
  printf("%s:\n", msg);
  for (i=0; i<4; i++)
    {
      for (j=0; j<4; j++)
	printf("%5.2lf ", mat[i*4+j] );
      printf("\n");
    }
}

/* ----------------------------------------------------------------------
 *
 * pomEulerVInvert - Invert pomEulerV. pomEulerV is overwritten.
 *
 * Returns : *POM_EULER_V or NULL if pomEulerV is not correct
 */

POM_EULER_V * pomEulerVInvert (POM_EULER_V *pomEulerV)
{
  T3D t3d;

  if (pomEulerVToT3D(pomEulerV, &t3d) == NULL)
    return NULL;

  if (t3dInvert(&t3d)  == NULL)
    return NULL;

  if (pomEulerVFromT3D(&t3d, pomEulerV) == NULL)
    return NULL;

  return pomEulerV;
}

/* ----------------------------------------------------------------------
 *
 * pomEulerVCompIn - Compose a with b, store the result in c and return c.
 *
 * Returns : *POM_EULER_V or NULL if pomEulerV is not correct
 */

POM_EULER_V * pomEulerVCompIn (POM_EULER_V *c, POM_EULER_V *a, POM_EULER_V *b)
{
  T3D ta, tb, tc = {{ .flags = T3D_ALLOW_CONVERSION }};

  if (pomEulerVToT3D(a, &ta) == NULL)
    return NULL;

  if (pomEulerVToT3D(b, &tb) == NULL)
    return NULL;

  if (t3dCompIn(&tc, &ta, &tb)  == NULL)
    return NULL;

  if (pomEulerVFromT3D(&tc, c) == NULL)
    return NULL;

  return c;
}


/* ----------------------------------------------------------------------
 *
 * pomEulerVComp - Compose a with b, store the result in a and return a.
 *
 * Returns : *POM_EULER_V or NULL if pomEulerV is not correct
 */

POM_EULER_V * pomEulerVComp (POM_EULER_V *a, POM_EULER_V *b)
{
  T3D ta, tb;

  if (pomEulerVToT3D(a, &ta) == NULL)
    return NULL;

  if (pomEulerVToT3D(b, &tb) == NULL)
    return NULL;

  if (t3dComp(&ta, &tb)  == NULL)
    return NULL;

  if (pomEulerVFromT3D(&ta, a) == NULL)
    return NULL;

  return a;
}


/* ----------------------------------------------------------------------
 *
 * pomEulerVToT3D - fillin t3d using pomEulerV
 *
 * Returns : *t3d or NULL if t3d is not correct
 */

T3D * pomEulerVToT3D (POM_EULER_V *pomEulerV, T3D *t3d)
{
  t3dSetOk(t3d);
  t3d->euler.flags = T3D_ALLOW_CONVERSION;
  t3d->type = T3D_EULER;

  /* fill in pos */
  t3d->euler.euler[0] = pomEulerV->euler.yaw;
  t3d->euler.euler[1] = pomEulerV->euler.pitch;
  t3d->euler.euler[2] = pomEulerV->euler.roll;
  t3d->euler.euler[3] = pomEulerV->euler.x;
  t3d->euler.euler[4] = pomEulerV->euler.y;
  t3d->euler.euler[5] = pomEulerV->euler.z;

  /* fill in variances */
  t3d->euler.v[0] = pomEulerV->var.voo;
  t3d->euler.v[1] = pomEulerV->var.vop;
  t3d->euler.v[2] = pomEulerV->var.vpp;
  t3d->euler.v[3] = pomEulerV->var.vor;
  t3d->euler.v[4] = pomEulerV->var.vpr;
  t3d->euler.v[5] = pomEulerV->var.vrr;
  t3d->euler.v[6] = pomEulerV->var.vox;
  t3d->euler.v[7] = pomEulerV->var.vpx;
  t3d->euler.v[8] = pomEulerV->var.vrx;
  t3d->euler.v[9] = pomEulerV->var.vxx;
  t3d->euler.v[10] = pomEulerV->var.voy;
  t3d->euler.v[11] = pomEulerV->var.vpy;
  t3d->euler.v[12] = pomEulerV->var.vry;
  t3d->euler.v[13] = pomEulerV->var.vxy;
  t3d->euler.v[14] = pomEulerV->var.vyy;
  t3d->euler.v[15] = pomEulerV->var.voz;
  t3d->euler.v[16] = pomEulerV->var.vpz;
  t3d->euler.v[17] = pomEulerV->var.vrz;
  t3d->euler.v[18] = pomEulerV->var.vxz;
  t3d->euler.v[19] = pomEulerV->var.vyz;
  t3d->euler.v[20] = pomEulerV->var.vzz;

  /* t3d converted to matrix */
  if (!t3dConvertTo(T3D_MATRIX,t3d))
    return NULL;

  return t3d;
}

/* ----------------------------------------------------------------------
 *
 * pomEulerVFromT3D - fillin pomEulerV using t3d
 *
 * Returns : *pomEulerV or NULL if t3d is not correct
 */

POM_EULER_V *pomEulerVFromT3D (T3D *t3d, POM_EULER_V *pomEulerV)
{

  /* t3d must be in T3D_EULER format */
  if (!t3dConvertTo(T3D_EULER,t3d))
    return NULL;

  /* fill in pos */
  pomEulerV->euler.yaw = t3d->euler.euler[0];
  pomEulerV->euler.pitch = t3d->euler.euler[1];
  pomEulerV->euler.roll = t3d->euler.euler[2];
  pomEulerV->euler.x = t3d->euler.euler[3];
  pomEulerV->euler.y = t3d->euler.euler[4];
  pomEulerV->euler.z = t3d->euler.euler[5];

  /* fill in variances */
  pomEulerV->var.voo = t3d->euler.v[0];
  pomEulerV->var.vop = t3d->euler.v[1];
  pomEulerV->var.vpp = t3d->euler.v[2];
  pomEulerV->var.vor = t3d->euler.v[3];
  pomEulerV->var.vpr = t3d->euler.v[4];
  pomEulerV->var.vrr = t3d->euler.v[5];
  pomEulerV->var.vox = t3d->euler.v[6];
  pomEulerV->var.vpx = t3d->euler.v[7];
  pomEulerV->var.vrx = t3d->euler.v[8];
  pomEulerV->var.vxx = t3d->euler.v[9];
  pomEulerV->var.voy = t3d->euler.v[10];
  pomEulerV->var.vpy = t3d->euler.v[11];
  pomEulerV->var.vry = t3d->euler.v[12];
  pomEulerV->var.vxy = t3d->euler.v[13];
  pomEulerV->var.vyy = t3d->euler.v[14];
  pomEulerV->var.voz = t3d->euler.v[15];
  pomEulerV->var.vpz = t3d->euler.v[16];
  pomEulerV->var.vrz = t3d->euler.v[17];
  pomEulerV->var.vxz = t3d->euler.v[18];
  pomEulerV->var.vyz = t3d->euler.v[19];
  pomEulerV->var.vzz = t3d->euler.v[20];

  return pomEulerV;
}

/* ----------------------------------------------------------------------
 *
 * pomEulerPointAtoB - Converts a point from frame A to frame B
 *                     
 * Returns : 
 */

void  pomEulerPointAtoB (POM_EULER *aToB, double pa[3], double pb[3])
{
  double aToBMat[16];

  t3dMatrixOfEuler((double *)aToB, aToBMat);

  pb[0] = pa[0]*aToBMat[0] + pa[1]*aToBMat[1] + pa[2]*aToBMat[2]  + aToBMat[3];
  pb[1] = pa[0]*aToBMat[4] + pa[1]*aToBMat[5] + pa[2]*aToBMat[6]  + aToBMat[7];
  pb[2] = pa[0]*aToBMat[8] + pa[1]*aToBMat[9] + pa[2]*aToBMat[10] + aToBMat[11];
}

/*----------------------------------------------------------------------*/

/* ascii file header */
static const char magic[] = "Pom sensor position\n";

/*
 * pomWriteSensorPos ----------------------------------------------------
 *
 * Write a PomSensorPos structure in a file
 */

int
pomWriteSensorPos(char *file, POM_SENSOR_POS *pos)
{
   FILE *f;

   if ((f = fopen(file, "w")) == NULL) {
      return -1;
   }

   fprintf(f, magic);
   fprintf(f, "date %d\nsensorToMain = ", pos->date);
   fprintf(f, "%f ", pos->sensorToMain.euler.yaw);
   fprintf(f, "%f ", pos->sensorToMain.euler.pitch);
   fprintf(f, "%f ", pos->sensorToMain.euler.roll);
   fprintf(f, "%f ", pos->sensorToMain.euler.x);
   fprintf(f, "%f ", pos->sensorToMain.euler.y);
   fprintf(f, "%f ", pos->sensorToMain.euler.z);
   fprintf(f, "\nmainToBase = ");
   fprintf(f, "%f ", pos->mainToBase.euler.yaw);
   fprintf(f, "%f ", pos->mainToBase.euler.pitch);
   fprintf(f, "%f ", pos->mainToBase.euler.roll);
   fprintf(f, "%f ", pos->mainToBase.euler.x);
   fprintf(f, "%f ", pos->mainToBase.euler.y);
   fprintf(f, "%f ", pos->mainToBase.euler.z);
   fprintf(f, "\nmainToOrigin = ");
   fprintf(f, "%f ", pos->mainToOrigin.euler.yaw);
   fprintf(f, "%f ", pos->mainToOrigin.euler.pitch);
   fprintf(f, "%f ", pos->mainToOrigin.euler.roll);
   fprintf(f, "%f ", pos->mainToOrigin.euler.x);
   fprintf(f, "%f ", pos->mainToOrigin.euler.y);
   fprintf(f, "%f ", pos->mainToOrigin.euler.z);
   fprintf(f, "\n");

   fclose(f);
   return 0;
}
/*
 * insituReadTag -------------------------------------------------------
 *
 * Read tag from a file
 */

int
pomReadSensorPos(char *file, POM_SENSOR_POS *pos)
{
   static char tmp[50];
   FILE *f;


   if ((f = fopen(file, "r")) == NULL) {
      return -1;
   }

   if (fgets(tmp, 50, f) == NULL) goto error;
   if (strcmp(tmp, magic)) goto error;

   /* initialize pos structure */
   memset(pos, 0, sizeof(*pos));

   /* read matrices - variances are lost! */
   if (fscanf(f, "date %d\n", &pos->date) != 1) goto error;
   if (fscanf(f, "sensorToMain = %lf %lf %lf %lf %lf %lf\n",
	      &pos->sensorToMain.euler.yaw,
	      &pos->sensorToMain.euler.pitch,
	      &pos->sensorToMain.euler.roll,
	      &pos->sensorToMain.euler.x,
	      &pos->sensorToMain.euler.y,
	      &pos->sensorToMain.euler.z) != 6) goto error;
   if (fscanf(f, "mainToBase = %lf %lf %lf %lf %lf %lf\n",
	      &pos->mainToBase.euler.yaw,
	      &pos->mainToBase.euler.pitch,
	      &pos->mainToBase.euler.roll,
	      &pos->mainToBase.euler.x,
	      &pos->mainToBase.euler.y,
	      &pos->mainToBase.euler.z) != 6) goto error;
   if (fscanf(f, "mainToOrigin = %lf %lf %lf %lf %lf %lf\n",
	      &pos->mainToOrigin.euler.yaw,
	      &pos->mainToOrigin.euler.pitch,
	      &pos->mainToOrigin.euler.roll,
	      &pos->mainToOrigin.euler.x,
	      &pos->mainToOrigin.euler.y,
	      &pos->mainToOrigin.euler.z) != 6) goto error;

   fclose(f);
   return 0;
  error:
   fclose(f);
   errno = EINVAL;
   return -1;
}
